import networkx as nx
import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as lg
from sklearn.preprocessing import normalize

from gem.embedding.static_graph_embedding import StaticGraphEmbedding


class LocallyLinearEmbedding(StaticGraphEmbedding):
    hyper_params = {
        'method_name': 'lle_svd'
    }

    def __init__(self, *args, **kwargs):
        """ Initialize the LocallyLinearEmbedding class

        Args:
            d: dimension of the embedding
        """
        super(LocallyLinearEmbedding, self).__init__(*args, **kwargs)

    def learn_embedding(self, graph=None,
                        is_weighted=False, no_python=False):
        if not graph:
            raise ValueError('graph needed')
        graph = graph.to_undirected()
        A = nx.to_scipy_sparse_matrix(graph)
        normalize(A, norm='l1', axis=1, copy=False)
        i_n = sp.eye(len(graph.nodes))
        i_min_A = i_n - A
        u, s, vt = lg.svds(i_min_A, k=self._d + 1, which='SM')
        self._X = vt.T
        self._X = self._X[:, 1:]
        return self._X.real

    def get_edge_weight(self, i, j):
        return np.exp(
            -np.power(np.linalg.norm(self._X[i, :] - self._X[j, :]), 2)
        )



import matplotlib
import matplotlib.pyplot as plt


def get_node_color(node_community):
    cnames = [item[0] for item in matplotlib.colors.cnames.iteritems()]
    node_colors = [cnames[c] for c in node_community]
    return node_colors


def plot(x_s, y_s, fig_n, x_lab, y_lab, file_save_path, title, legendLabels=None, show=False):
    plt.rcParams.update({'font.size': 16, 'font.weight': 'bold'})
    markers = ['o', '*', 'v', 'D', '<' , 's', '+', '^', '>']
    colors = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
    series = []
    plt.figure(fig_n)
    for i in range(len(x_s)):
        x = x_s[i]
        y = y_s[i]
        series.append(plt.plot(x, y, color=colors[i], linewidth=2, marker=markers[i], markersize=8))
        plt.xlabel(x_lab, fontsize=16, fontweight='bold')
        plt.ylabel(y_lab, fontsize=16, fontweight='bold')
        plt.title(title, fontsize=16, fontweight='bold')
    if legendLabels:
        plt.legend([s[0] for s in series], legendLabels)
    plt.savefig(file_save_path)
    if show:
        plt.show()


def plot_ts(ts_df, plot_title, eventDates, eventLabels=None, save_file_name=None, xLabel=None, yLabel=None, show=False):
    ax = ts_df.plot(title=plot_title, marker='*', markerfacecolor='red', markersize=10, linestyle='solid')
    colors = ['r', 'g', 'c', 'm', 'y', 'b', 'k']
    if not eventLabels:
        for eventDate in eventDates:
            ax.axvline(eventDate, color='r', linestyle='--', lw=2)  # Show event as a red vertical line
    else:
        for idx in range(len(eventDates)):
            # Show event as a red vertical line
            ax.axvline(eventDates[idx], color=colors[idx], linestyle='--', lw=2, label=eventLabels[idx])
            ax.legend()
    if xLabel:
        ax.set_xlabel(xLabel, fontweight='bold')
    if yLabel:
        ax.set_ylabel(yLabel, fontweight='bold')
    fig = ax.get_figure()
    if save_file_name:
        fig.savefig(save_file_name, bbox_inches='tight')
    if show:
        fig.show()



import unittest

from gem.embedding.gf import GraphFactorization
from gem.embedding.hope import HOPE
from gem.embedding.lap import LaplacianEigenmaps
from gem.embedding.lle import LocallyLinearEmbedding
from gem.embedding.node2vec import node2vec
from gem.embedding.sdne import SDNE


class EmbeddingsTest(unittest.TestCase):

    def test_hope(self):
        model = HOPE()
        self._run(model)

    def test_le(self):
        model = LaplacianEigenmaps()
        self._run(model)

    def test_lle(self):
        model = LocallyLinearEmbedding()
        self._run(model)

    def test_node2vec(self):
        model = node2vec()
        self._run(model)

    def test_sdne(self):
        model = SDNE()
        self._run(model)

    def test_gf(self):
        model = GraphFactorization()
        self._run(model)

    def _run(self, model):
        try:
            model.learn_embedding()
        except ValueError:
            pass
        # check method name
        self.assertEqual(model.hyper_params['method_name'], model.get_method_name())
        # check unlearned prediction
        try:
            model.get_embedding()
        except ValueError:
            pass



import numpy as np

from tensorflow.keras.layers import Input, Dense
from tensorflow.keras.models import Model, model_from_json
import tensorflow.keras.regularizers as Reg


def model_batch_predictor(model, X, batch_size):
    n_samples = X.shape[0]
    counter = 0
    pred = None
    while counter < n_samples // batch_size:
        _, curr_pred = \
            model.predict(X[batch_size * counter:batch_size * (counter + 1),
                          :].toarray())
        if counter:
            pred = np.vstack((pred, curr_pred))
        else:
            pred = curr_pred
        counter += 1
    if n_samples % batch_size != 0:
        _, curr_pred = \
            model.predict(X[batch_size * counter:, :].toarray())
        if counter:
            pred = np.vstack((pred, curr_pred))
        else:
            pred = curr_pred
    return pred


def batch_generator_sdne(X, beta, batch_size, shuffle):
    row_indices, col_indices = X.nonzero()
    sample_index = np.arange(row_indices.shape[0])
    number_of_batches = row_indices.shape[0] // batch_size
    counter = 0
    if shuffle:
        np.random.shuffle(sample_index)
    while True:
        batch_index = \
            sample_index[batch_size * counter:batch_size * (counter + 1)]
        X_batch_v_i = X[row_indices[batch_index], :].toarray()
        X_batch_v_j = X[col_indices[batch_index], :].toarray()
        InData = np.append(X_batch_v_i, X_batch_v_j, axis=1)

        B_i = np.ones(X_batch_v_i.shape)
        B_i[X_batch_v_i != 0] = beta
        B_j = np.ones(X_batch_v_j.shape)
        B_j[X_batch_v_j != 0] = beta
        X_ij = X[row_indices[batch_index], col_indices[batch_index]]
        deg_i = np.sum(X_batch_v_i != 0, 1).reshape((batch_size, 1))
        deg_j = np.sum(X_batch_v_j != 0, 1).reshape((batch_size, 1))
        a1 = np.append(B_i, deg_i, axis=1)
        a2 = np.append(B_j, deg_j, axis=1)
        OutData = [a1, a2, X_ij.T]
        counter += 1
        yield InData, OutData
        if counter == number_of_batches:
            if shuffle:
                np.random.shuffle(sample_index)
            counter = 0


def get_encoder(node_num, d, K, n_units, nu1, nu2, activation_fn):
    # Input
    x = Input(shape=(node_num,))
    # Encoder layers
    y = [None] * (K + 1)
    y[0] = x  # y[0] is assigned the input
    for i in range(K - 1):
        y[i + 1] = Dense(n_units[i],
                         activation=activation_fn,
                         kernel_regularizer=Reg.l1_l2(l1=nu1, l2=nu2))(y[i])
    y[K] = Dense(d,
                 activation=activation_fn,
                 kernel_regularizer=Reg.l1_l2(l1=nu1, l2=nu2))(y[K - 1])
    # Encoder model
    encoder = Model(inputs=x, outputs=y[K])
    return encoder


def get_decoder(node_num, d, K,
                n_units, nu1, nu2,
                activation_fn):
    # Input
    y = Input(shape=(d,))
    # Decoder layers
    y_hat = [None] * (K + 1)
    y_hat[K] = y
    for i in range(K - 1, 0, -1):
        y_hat[i] = Dense(n_units[i - 1],
                         activation=activation_fn,
                         kernel_regularizer=Reg.l1_l2(l1=nu1, l2=nu2))(y_hat[i + 1])
    y_hat[0] = Dense(node_num, activation=activation_fn,
                     kernel_regularizer=Reg.l1_l2(l1=nu1, l2=nu2))(y_hat[1])
    # Output
    x_hat = y_hat[0]  # decoder's output is also the actual output
    # Decoder Model
    decoder = Model(inputs=y, outputs=x_hat)
    return decoder


def get_autoencoder(encoder, decoder):
    # Input
    x = Input(shape=(encoder.layers[0].input_shape[0][1],))
    # Generate embedding
    y = encoder(x)
    # Generate reconstruction
    x_hat = decoder(y)
    # Autoencoder Model
    autoencoder = Model(inputs=x, outputs=[x_hat, y])
    return autoencoder


def graphify(reconstruction):
    [n1, n2] = reconstruction.shape
    n = min(n1, n2)
    reconstruction = np.copy(reconstruction[0:n, 0:n])
    reconstruction = (reconstruction + reconstruction.T) / 2
    reconstruction -= np.diag(np.diag(reconstruction))
    return reconstruction


def loadmodel(filename):
    model = None
    try:
        model = model_from_json(open(filename).read())
    except (FileNotFoundError, IOError):
        print('Error reading file: {0}. Cannot load previous model'.format(filename))
        exit()
    return model


def loadweights(model, filename):
    try:
        model.load_weights(filename)
    except (FileNotFoundError, IOError):
        print('Error reading file: {0}. Cannot load previous weights'.format(filename))
        exit()


def savemodel(model, filename):
    json_string = model.to_json()
    open(filename, 'w').write(json_string)


def saveweights(model, filename):
    model.save_weights(filename, overwrite=True)



import os
import sys
import numpy as np
from subprocess import call
import matplotlib.pyplot as plt

from gem.embedding.static_graph_embedding import StaticGraphEmbedding
from gem.utils import graph_util


class GraphFactorization(StaticGraphEmbedding):

    """`Graph Factorization`_.
    Graph Factorization factorizes the adjacency matrix with regularization.

    Args:
        hyper_dict (object): Hyper parameters.
        kwargs (dict): keyword arguments, form updating the parameters

    Examples:
        >>> from gem.embedding.gf import GraphFactorization
        >>> edge_f = 'data/karate.edgelist'
        >>> G = graph_util.loadGraphFromEdgeListTxt(edge_f, directed=False)
        >>> G = G.to_directed()
        >>> res_pre = 'results/testKarate'
        >>> graph_util.print_graph_stats(G)
        >>> t1 = time()
        >>> embedding = GraphFactorization(2, 100000, 1 * 10**-4, 1.0)
        >>> embedding.learn_embedding(graph=G, edge_f=None,
                                  is_weighted=True, no_python=True)
        >>> print ('Graph Factorization:Training time: %f' % (time() - t1))
        >>> viz.plot_embedding2D(embedding.get_embedding(),
                             di_graph=G, node_colors=None)
        >>> plt.show()
    .. _Graph Factorization:
        https://static.googleusercontent.com/media/research.google.com/en//pubs/archive/40839.pdf
    """
    hyper_params = {
        'print_step': 10000,
        'method_name': 'graph_factor_sgd'
    }

    def __init__(self, *args, **kwargs):
        """ Initialize the GraphFactorization class
        Args:
            d: dimension of the embedding
            eta: learning rate of sgd
            regu: regularization coefficient of magnitude of weights
            max_iter: max iterations in sgd
            print_step: #iterations to log the prgoress (step%print_step)
        """
        super(GraphFactorization, self).__init__(*args, **kwargs)

    def _use_c_implementation(self, graph=None):
        if sys.platform[0] == "w":  # pragma: no cover
            args = ["gem/c_exe/gf.exe"]
        else:
            args = ["gem/c_exe/gf"]
        os.makedirs('gem/intermediate', exist_ok=True)
        graph_filename = 'gem/intermediate/%s_gf.graph' % self._data_set
        emb_filename = 'gem/intermediate/%s_%d_gf.emb' % (self._data_set, self._d)
        graph_util.saveGraphToEdgeListTxt(graph, graph_filename)
        args.append(graph_filename)
        args.append(emb_filename)
        args.append("1")  # Verbose
        args.append("1")  # Weighted
        args.append("%d" % self._d)
        args.append("%f" % self._eta)
        args.append("%f" % self._regu)
        args.append("%d" % self._max_iter)
        args.append("%d" % self._print_step)
        call(args)
        # this should be tested in future
        try:  # pragma: no cover
            self._X = graph_util.loadEmbedding(emb_filename)
        except FileNotFoundError:
            self._X = np.random.randn(len(graph.nodes), self._d)
        os.remove(emb_filename)
        return self._X

    def learn_embedding(self, graph=None,
                        is_weighted=False, no_python=True):
        if not graph:
            raise ValueError('graph needed')
        if no_python:
            try:
                self._use_c_implementation(graph)
            except FileNotFoundError:
                print('./gf not found. Reverting to Python implementation. Please compile gf, place node2vec in '
                      'the path and grant executable permission')
        self._node_num = len(graph.nodes)
        self._X = 0.01 * np.random.randn(self._node_num, self._d)
        for _ in range(self._max_iter):
            for i, j, w in graph.edges(data='weight', default=1):
                if j <= i:
                    continue
                term1 = -(w - np.dot(self._X[i, :], self._X[j, :])) * self._X[j, :]
                term2 = self._regu * self._X[i, :]
                del_phi = term1 + term2
                self._X[i, :] -= self._eta * del_phi
        return self._X

    def get_edge_weight(self, i, j):
        return np.dot(self._X[i, :], self._X[j, :])



import numpy as np
import networkx as nx

from tensorflow.keras.layers import Input, Lambda, Subtract
from tensorflow.keras.models import Model, model_from_json
from tensorflow.keras.optimizers import SGD
from tensorflow.keras import backend as KBack

from gem.embedding.static_graph_embedding import StaticGraphEmbedding
from gem.embedding.sdne_utils import get_encoder, get_decoder, get_autoencoder, batch_generator_sdne, \
    model_batch_predictor, saveweights, savemodel, graphify


class SDNE(StaticGraphEmbedding):
    hyper_params = {
        'method_name': 'sdne',
        'actfn': 'relu',
        'modelfile': None,
        'weightfile': None,
        'savefilesuffix': None

    }

    def __init__(self, *args, **kwargs):
        """ Initialize the SDNE class

        Args:
            d: dimension of the embedding
            beta: penalty parameter in matrix B of 2nd order objective
            alpha: weighing hyperparameter for 1st order objective
            nu1: L1-reg hyperparameter
            nu2: L2-reg hyperparameter
            K: number of hidden layers in encoder/decoder
            n_units: vector of length K-1 containing #units in hidden layers
                     of encoder/decoder, not including the units in the
                     embedding layer
            rho: bounding ratio for number of units in consecutive layers (< 1)
            n_iter: number of sgd iterations for first embedding (const)
            xeta: sgd step size parameter
            n_batch: minibatch size for SGD
            modelfile: Files containing previous encoder and decoder models
            weightfile: Files containing previous encoder and decoder weights
        """
        super(SDNE, self).__init__(*args, **kwargs)
        self.learned = False

    def learn_embedding(self, graph=None,
                        is_weighted=False, no_python=False):
        if not graph:
            raise ValueError('graph needed')
        sparse = nx.to_scipy_sparse_matrix(graph)
        sparse = (sparse + sparse.T) / 2
        self._node_num = len(graph.nodes)

        # Generate encoder, decoder and autoencoder
        self._num_iter = self._n_iter
        # If cannot use previous step information, initialize new models
        self._encoder = get_encoder(self._node_num, self._d,
                                    self._K, self._n_units,
                                    self._nu1, self._nu2,
                                    self._actfn)
        self._decoder = get_decoder(self._node_num, self._d,
                                    self._K, self._n_units,
                                    self._nu1, self._nu2,
                                    self._actfn)
        self._autoencoder = get_autoencoder(self._encoder, self._decoder)

        # Initialize self._model
        # Input
        x_in = Input(shape=(2 * self._node_num,), name='x_in')
        x1 = Lambda(
            lambda x: x[:, 0:self._node_num],
            output_shape=(self._node_num,)
        )(x_in)
        x2 = Lambda(
            lambda x: x[:, self._node_num:2 * self._node_num],
            output_shape=(self._node_num,)
        )(x_in)
        # Process inputs
        [x_hat1, y1] = self._autoencoder(x1)
        [x_hat2, y2] = self._autoencoder(x2)
        # Outputs
        x_diff1 = Subtract()([x_hat1, x1])
        x_diff2 = Subtract()([x_hat2, x2])
        y_diff = Subtract()([y2, y1])

        # Objectives
        def weighted_mse_x(y_true, y_pred):
            """ Hack: This fn doesn't accept additional arguments.
                      We use y_true to pass them.
                y_pred: Contains x_hat - x
                y_true: Contains [b, deg]
            """
            return KBack.sum(
                KBack.square(y_pred * y_true[:, 0:self._node_num]),
                axis=-1) / y_true[:, self._node_num]

        def weighted_mse_y(y_true, y_pred):
            """ Hack: This fn doesn't accept additional arguments.
                      We use y_true to pass them.
            y_pred: Contains y2 - y1
            y_true: Contains s12
            """
            min_batch_size = KBack.shape(y_true)[0]
            return KBack.reshape(
                KBack.sum(KBack.square(y_pred), axis=-1),
                [min_batch_size, 1]
            ) * y_true

        # Model
        self._model = Model(inputs=x_in, outputs=[x_diff1, x_diff2, y_diff])
        sgd = SGD(lr=self._xeta, decay=1e-5, momentum=0.99, nesterov=True)
        self._model.compile(
            optimizer=sgd,
            loss=[weighted_mse_x, weighted_mse_x, weighted_mse_y],
            loss_weights=[1, 1, self._alpha]
        )

        self._model.fit_generator(
            generator=batch_generator_sdne(sparse, self._beta, self._n_batch, True),
            epochs=self._num_iter,
            steps_per_epoch=sparse.nonzero()[0].shape[0] // self._n_batch,
            verbose=1
        )
        # Get embedding for all points
        self._Y = model_batch_predictor(self._autoencoder, sparse, self._n_batch)
        # Save the autoencoder and its weights
        if self._weightfile is not None:
            saveweights(self._encoder, self._weightfile[0])
            saveweights(self._decoder, self._weightfile[1])
        if self._modelfile is not None:
            savemodel(self._encoder, self._modelfile[0])
            savemodel(self._decoder, self._modelfile[1])
        if self._savefilesuffix is not None:
            saveweights(
                self._encoder,
                'encoder_weights_' + self._savefilesuffix + '.hdf5'
            )
            saveweights(
                self._decoder,
                'decoder_weights_' + self._savefilesuffix + '.hdf5'
            )
            savemodel(
                self._encoder,
                'encoder_model_' + self._savefilesuffix + '.json'
            )
            savemodel(
                self._decoder,
                'decoder_model_' + self._savefilesuffix + '.json'
            )
            # Save the embedding
            np.savetxt('embedding_' + self._savefilesuffix + '.txt', self._Y)
        self.learned = True
        return self._Y

    def get_embedding(self, filesuffix=None):
        if not self.learned:
            raise ValueError("Embedding not learned yet")
        return self._Y if filesuffix is None else np.loadtxt(
            'embedding_' + filesuffix + '.txt'
        )

    def get_edge_weight(self, i, j, embed=None, filesuffix=None):
        if embed is None:
            if filesuffix is None:
                embed = self._Y
            else:
                embed = np.loadtxt('embedding_' + filesuffix + '.txt')
        if i == j:
            return 0
        else:
            sparse_hat = self.get_reconst_from_embed(embed[(i, j), :], filesuffix)
            return (sparse_hat[i, j] + sparse_hat[j, i]) / 2

    def get_reconstructed_adj(self, embed=None, node_l=None, filesuffix=None):
        if embed is None:
            if filesuffix is None:
                embed = self._Y
            else:
                embed = np.loadtxt('embedding_' + filesuffix + '.txt')
        sparse_hat = self.get_reconst_from_embed(embed, node_l, filesuffix)
        return graphify(sparse_hat)

    def get_reconst_from_embed(self, embed, node_l=None, filesuffix=None):
        if filesuffix is None:
            if node_l is not None:
                return self._decoder.predict(
                    embed,
                    batch_size=self._n_batch)[:, node_l]
            else:
                return self._decoder.predict(embed, batch_size=self._n_batch)
        else:
            try:
                decoder = model_from_json(
                    open('decoder_model_' + filesuffix + '.json').read()
                )
            except FileNotFoundError:
                print('Error reading file: {0}. Cannot load previous model'.format('decoder_model_'+filesuffix+'.json'))
                exit()
            try:
                decoder.load_weights('decoder_weights_' + filesuffix + '.hdf5')
            except (FileNotFoundError, ReferenceError):
                print('Error reading file: {0}. Cannot load previous weights'.format('decoder_weights_'+filesuffix+'.hdf5'))
                exit()
            if node_l is not None:
                return decoder.predict(embed, batch_size=self._n_batch)[:, node_l]
            else:
                return decoder.predict(embed, batch_size=self._n_batch)



import networkx as nx
import numpy as np
import scipy.sparse.linalg as lg
from gem.embedding.static_graph_embedding import StaticGraphEmbedding
from gem.utils import graph_util


class HOPE(StaticGraphEmbedding):

    hyper_params = {
        'method_name': 'hope_gsvd'
    }

    def __init__(self, *args, **kwargs):
        """ Initialize the HOPE class

        Args:
            d: dimension of the embedding
            beta: higher order coefficient
        """
        super(HOPE, self).__init__(*args, **kwargs)

    def learn_embedding(self, graph=None,
                        is_weighted=False, no_python=False):
        if not graph:
            raise ValueError('graph needed')

        A = nx.to_numpy_matrix(graph)
        m_g = np.eye(len(graph.nodes)) - self._beta * A
        m_l = self._beta * A
        S = np.dot(np.linalg.inv(m_g), m_l)

        u, s, vt = lg.svds(S, k=self._d // 2)
        X1 = np.dot(u, np.diag(np.sqrt(s)))
        X2 = np.dot(vt.T, np.diag(np.sqrt(s)))
        self._X = np.concatenate((X1, X2), axis=1)

        p_d_p_t = np.dot(u, np.dot(np.diag(s), vt))
        eig_err = np.linalg.norm(p_d_p_t - S)
        print('SVD error (low rank): %f' % eig_err)
        return self._X

    def get_edge_weight(self, i, j):
        return np.dot(self._X[i, :self._d // 2], self._X[j, self._d // 2:])



import numpy as np
import secrets


def get_random_edge_pairs(node_num, sample_ratio=0.01, is_undirected=True):
    num_pairs = int(sample_ratio * node_num * (node_num - 1))
    if is_undirected:
        num_pairs = num_pairs / 2
    current_sets = set()
    while len(current_sets) < num_pairs:
        p = (secrets.randbelow(node_num), secrets.randbelow(node_num))
        if p in current_sets:
            continue
        if is_undirected and (p[1], p[0]) in current_sets:
            continue
        current_sets.add(p)
    return list(current_sets)


def get_edge_list_from_adj_mtrx(adj, threshold=0.0, is_undirected=True, edge_pairs=None):
    result = []
    node_num = adj.shape[0]
    if edge_pairs:
        for (st, ed) in edge_pairs:
            if adj[st, ed] >= threshold:
                result.append((st, ed, adj[st, ed]))
    else:
        for i in range(node_num):
            for j in range(node_num):
                if j == i:
                    continue
                if is_undirected and i >= j:
                    continue
                if adj[i, j] > threshold:
                    result.append((i, j, adj[i, j]))
    return result


def split_di_graph_to_train_test(di_graph, train_ratio, is_undirected=True):
    train_digraph = di_graph.copy()
    test_digraph = di_graph.copy()
    for (st, ed, w) in list(di_graph.edges(data='weight', default=1)):
        if is_undirected and st >= ed:
            continue
        if np.random.uniform() <= train_ratio:
            test_digraph.remove_edge(st, ed)
            if is_undirected:
                test_digraph.remove_edge(ed, st)
        else:
            train_digraph.remove_edge(st, ed)
            if is_undirected:
                train_digraph.remove_edge(ed, st) 
    return train_digraph, test_digraph



import numpy as np

precision_pos = [2, 10, 100, 200, 300, 500, 1000]


def computePrecisionCurve(predicted_edge_list, true_digraph, max_k=-1):
    if max_k == -1:
        max_k = len(predicted_edge_list)
    else:
        max_k = min(max_k, len(predicted_edge_list))

    sorted_edges = sorted(predicted_edge_list, key=lambda x: x[2], reverse=True)

    precision_scores = []
    delta_factors = []
    correct_edge = 0
    for i in range(max_k):
        if true_digraph.has_edge(sorted_edges[i][0], sorted_edges[i][1]):
            correct_edge += 1
            delta_factors.append(1.0)
        else:
            delta_factors.append(0.0)
        precision_scores.append(1.0 * correct_edge / (i + 1))
    return precision_scores, delta_factors


def computeMAP(predicted_edge_list, true_digraph, max_k=-1, is_undirected=False):
    node_num = len(true_digraph.nodes)
    node_edges = []
    for i in range(node_num):
        node_edges.append([])
    for (st, ed, w) in predicted_edge_list:
        node_edges[st].append((st, ed, w))
    node_ap = [0.0] * node_num
    count = 0
    for i in range(node_num):
        if not is_undirected and true_digraph.out_degree(i) == 0:
            continue
        count += 1
        precision_scores, delta_factors = computePrecisionCurve(node_edges[i], true_digraph, max_k)
        precision_rectified = [p * d for p, d in zip(precision_scores, delta_factors)]
        if sum(delta_factors) == 0:
            node_ap[i] = 0
        else:
            node_ap[i] = float(sum(precision_rectified) / sum(delta_factors))
    return sum(node_ap) / count



from gem.evaluation import evaluate_graph_reconstruction as gr


def fit_model(graph, model, verbose: bool = False):
    if verbose:
        print('Num nodes: %d, num edges: %d' % (graph.number_of_nodes(), graph.number_of_edges()))
    # Learn embedding - accepts a networkx graph or file with edge list
    y = model.learn_embedding(graph=graph, is_weighted=True, no_python=True)
    # Evaluate on graph reconstruction
    return gr.evaluateStaticGraphReconstruction(graph, model, y, None)



import networkx as nx
import numpy as np
import scipy.sparse.linalg as lg

from gem.embedding.static_graph_embedding import StaticGraphEmbedding


class LaplacianEigenmaps(StaticGraphEmbedding):
    hyper_params = {
        'method_name': 'lap_eigmap_svd'
    }

    def __init__(self, *args, **kwargs):
        """ Initialize the LaplacianEigenmaps class

        Args:
            d: dimension of the embedding
        """
        super(LaplacianEigenmaps, self).__init__(*args, **kwargs)

    def learn_embedding(self, graph=None,
                        is_weighted=False, no_python=False):
        if not graph:
            raise ValueError('graph needed')
        graph = graph.to_undirected()
        l_sym = nx.normalized_laplacian_matrix(graph)

        w, v = lg.eigs(l_sym, k=self._d + 1, which='SM')
        idx = np.argsort(w)  # sort eigenvalues
        w = w[idx]
        v = v[:, idx]
        self._X = v[:, 1:]

        p_d_p_t = np.dot(v, np.dot(np.diag(w), v.T))
        eig_err = np.linalg.norm(p_d_p_t - l_sym)
        print('Laplacian matrix recon. error (low rank): %f' % eig_err)
        return self._X.real

    def get_edge_weight(self, i, j):
        return np.exp(
            -np.power(np.linalg.norm(self._X[i, :] - self._X[j, :]), 2)
        )



import matplotlib.pyplot as plt
import networkx as nx
from sklearn.manifold import TSNE
from gem.utils import plot_util


def plot_embedding2D(node_pos, node_colors=None, di_graph=None, labels=None):
    node_num, embedding_dimension = node_pos.shape
    if embedding_dimension > 2:
        print("Embedding dimension greater than 2, use tSNE to reduce it to 2")
        model = TSNE(n_components=2)
        node_pos = model.fit_transform(node_pos)

    if di_graph is None:
        # plot using plt scatter
        plt.scatter(node_pos[:, 0], node_pos[:, 1], c=node_colors)
    else:
        # plot using networkx with edge structure
        pos = {}
        for i in range(node_num):
            pos[i] = node_pos[i, :]
        if node_colors is not None:
            nx.draw_networkx_nodes(di_graph, pos,
                                   node_color=node_colors,
                                   width=0.1, node_size=100,
                                   arrows=False, alpha=0.8,
                                   font_size=5, labels=labels)
        else:
            nx.draw_networkx(di_graph, pos, node_color=node_colors,
                             width=0.1, node_size=300, arrows=False,
                             alpha=0.8, font_size=12, labels=labels)



"""
Run the graph embedding methods on Karate graph and evaluate them on
graph reconstruction and visualization. Please copy the
gem/data/karate.edgelist to the working directory
"""
import os.path
import unittest

import networkx as nx
import pickle
import numpy as np

from gem.embedding.gf import GraphFactorization
from gem.embedding.hope import HOPE
from gem.embedding.lap import LaplacianEigenmaps
from gem.embedding.lle import LocallyLinearEmbedding
from gem.embedding.node2vec import node2vec
from gem.embedding.sdne import SDNE


from tests.fit_model import fit_model


class SBMTest(unittest.TestCase):

    def setUp(self) -> None:
        # File that contains the edges. Format: source target
        # Optionally, you can add weights as third column: source target weight
        self.source_dir = os.path.dirname(os.path.abspath(__file__))
        file_prefix = os.path.join(self.source_dir, 'data/sbm.gpickle')

        # Load graph
        G = nx.read_gpickle(file_prefix)
        # convert G (networkx 1.x digraph) to networkx 2.x
        H = nx.DiGraph()
        H.add_nodes_from(G.node)
        for source_node in G.edge.keys():
            for target_node in G.edge[source_node].keys():
                H.add_edge(source_node, target_node)
        G = H
        try:
            node_colors = pickle.load(
                open(os.path.join(self.source_dir, 'data/sbm_node_labels.pickle'), 'rb')
            )
        except UnicodeDecodeError:
            node_colors = pickle.load(
                open(os.path.join(self.source_dir, 'data/sbm_node_labels.pickle'), 'rb'), encoding='latin1'
            )
        node_colors_arr = [None] * node_colors.shape[0]
        for idx in range(node_colors.shape[0]):
            node_colors_arr[idx] = np.where(node_colors[idx, :].toarray() == 1)[1][0]

        self.node_colors_arr = node_colors_arr
        self.G = G

    def test_GraphFactorization(self):
        model = GraphFactorization(d=128, max_iter=1000, eta=1 * 10**-4, regu=1.0, data_set='sbm')
        target = np.loadtxt(os.path.join(self.source_dir, 'smb_res/GraphFactorization.txt'))
        self.internal_model_test(model, target)

    def test_HOPE(self):
        model = HOPE(d=256, beta=0.01)
        target = np.loadtxt(os.path.join(self.source_dir, 'smb_res/HOPE.txt'))
        self.internal_model_test(model, target)

    def test_LaplacianEigenmaps(self):
        model = LaplacianEigenmaps(d=128)
        target = np.loadtxt(os.path.join(self.source_dir, 'smb_res/LaplacianEigenmaps.txt'))
        self.internal_model_test(model, target)

    def test_LocallyLinearEmbedding(self):
        model = LocallyLinearEmbedding(d=128)
        target = np.loadtxt(os.path.join(self.source_dir, 'smb_res/LocallyLinearEmbedding.txt'))
        self.internal_model_test(model, target)

    def test_node2vec(self):
        model = node2vec(d=182, max_iter=1, walk_len=80, num_walks=10, con_size=10, ret_p=1, inout_p=1, data_set='sbm')
        target = np.loadtxt(os.path.join(self.source_dir, 'smb_res/node2vec.txt'))
        self.internal_model_test(model, target, delta=.1)

    def test_SDNE(self):
        model = SDNE(d=128, beta=5, alpha=1e-5, nu1=1e-6, nu2=1e-6, K=3, n_units=[500, 300, ], rho=0.3, n_iter=30,
                     xeta=0.001, n_batch=500, modelfile=['enc_model.json', 'dec_model.json'],
                     weightfile=['enc_weights.hdf5', 'dec_weights.hdf5'])
        target = np.loadtxt(os.path.join(self.source_dir, 'smb_res/SDNE.txt'))
        self.internal_model_test(model, target, delta=1)

    def internal_model_test(self, model, target, verbose: bool = False, delta: float = 1e-3):
        MAP, prec_curv, err, err_baseline = fit_model(self.G, model)
        # ---------------------------------------------------------------------------------
        if verbose:
            print(("\tMAP: {} \t preccision curve: {}\n\n\n\n" + '-' * 100).format(MAP, prec_curv[:5]))

        self.assertTrue(abs(np.mean(target - model.get_embedding())) < delta)



import pickle
from gem.evaluation import metrics
from gem.utils import evaluation_util, graph_util
import networkx as nx
import numpy as np


def evaluateStaticGraphReconstruction(digraph, graph_embedding,
                                      X_stat, node_l=None, file_suffix=None,
                                      sample_ratio_e=None, is_undirected=True,
                                      is_weighted=False):
    node_num = len(digraph.nodes)
    # evaluation
    if sample_ratio_e:
        eval_edge_pairs = evaluation_util.get_random_edge_pairs(
            node_num,
            sample_ratio_e,
            is_undirected
        )
    else:
        eval_edge_pairs = None
    if file_suffix is None:
        estimated_adj = graph_embedding.get_reconstructed_adj(X_stat, node_l)
    else:
        estimated_adj = graph_embedding.get_reconstructed_adj(
            X_stat,
            file_suffix,
            node_l
        )
    predicted_edge_list = evaluation_util.get_edge_list_from_adj_mtrx(
        estimated_adj,
        is_undirected=is_undirected,
        edge_pairs=eval_edge_pairs
    )
    MAP = metrics.computeMAP(predicted_edge_list, digraph, is_undirected=is_undirected)
    prec_curv, _ = metrics.computePrecisionCurve(predicted_edge_list, digraph)
    # If weighted, compute the error in reconstructed weights of observed edges
    if is_weighted:
        digraph_adj = nx.to_numpy_matrix(digraph)
        estimated_adj[digraph_adj == 0] = 0
        err = np.linalg.norm(digraph_adj - estimated_adj)
        err_baseline = np.linalg.norm(digraph_adj)
    else:
        err = None
        err_baseline = None
    return MAP, prec_curv, err, err_baseline



try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages



VERSION = '2.0.0'
DISTNAME = 'nxt_gem'
MAINTAINER = 'Jan Ernsting'
MAINTAINER_EMAIL = 'j.ernsting@uni-muenster.de'
DESCRIPTION = 'nxt_gem: A Python module for Graph Embedding Methods'
LONG_DESCRIPTION = open('README.md').read()
URL = 'https://github.com/jernsting/nxt_gem'
DOWNLOAD_URL = 'https://github.com/jernsting/nxt_gem/archive/' + VERSION + '.tar.gz'
KEYWORDS = ['graph embedding', 'network analysis',
            'network embedding', 'data mining', 'machine learning']
LICENSE = 'BSD'
ISRELEASED = True

INSTALL_REQUIRES = (
    'numpy>=1.12.0',
    'scipy>=0.19.0',
    'networkx>=2.4',
    'matplotlib>=2.0.0',
    'scikit-learn>=0.21.2',
    'theano>=0.9.0',
)


def setup_package():
    setup(
        name=DISTNAME,
        version=VERSION,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        url=URL,
        download_url=DOWNLOAD_URL,
        keywords=KEYWORDS,
        install_requires=INSTALL_REQUIRES,
        packages=find_packages(),
        include_package_data=True,
        license=LICENSE,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        classifiers=['Intended Audience :: Science/Research',
                     'Intended Audience :: Developers',
                     'License :: OSI Approved',
                     'Programming Language :: Python',
                     'Topic :: Software Development',
                     'Topic :: Scientific/Engineering',
                     'Topic :: Scientific/Engineering :: Artificial Intelligence',
                     'Programming Language :: Python :: 3', ],
        )


if __name__ == "__main__":
    setup_package()



import os
import numpy as np
from subprocess import call
from gem.embedding.static_graph_embedding import StaticGraphEmbedding
from gem.utils import graph_util


class node2vec(StaticGraphEmbedding):
    hyper_params = {
        'method_name': 'node2vec_rw'
    }

    def __init__(self, *args, **kwargs):
        """ Initialize the node2vec class

        Args:
            d: dimension of the embedding
            max_iter: max iterations
            walk_len: length of random walk
            num_walks: number of random walks
            con_size: context size
            ret_p: return weight
            inout_p: inout weight
        """
        super(node2vec, self).__init__(*args, **kwargs)

    def learn_embedding(self, graph=None,
                        is_weighted=False, no_python=False):
        current_dir = os.path.dirname(os.path.abspath(__file__))
        executable = os.path.abspath(os.path.join(current_dir, '../c_exe/node2vec'))
        args = [executable]
        if not graph:
            raise ValueError('graph needed')
        graph_util.saveGraphToEdgeListTxtn2v(graph, 'tempGraph.graph')
        args.append("-i:tempGraph.graph")
        args.append("-o:tempGraph.emb")
        args.append("-d:%d" % self._d)
        args.append("-l:%d" % self._walk_len)
        args.append("-r:%d" % self._num_walks)
        args.append("-k:%d" % self._con_size)
        args.append("-e:%d" % self._max_iter)
        args.append("-p:%f" % self._ret_p)
        args.append("-q:%f" % self._inout_p)
        args.append("-v")
        args.append("-dr")
        args.append("-w")
        try:
            call(args)
        except Exception as e:  # pragma: no cover
            print(str(e))
            raise FileNotFoundError('./node2vec not found. Please compile snap, place node2vec in the system path '
                                    'and grant executable permission')
        self._X = graph_util.loadEmbedding('tempGraph.emb')
        return self._X

    def get_edge_weight(self, i, j):
        return np.dot(self._X[i, :], self._X[j, :])



import pickle
import numpy as np
import networkx as nx
import random
import itertools
import time


def transform_DiGraph_to_adj(di_graph):
    n = len(di_graph.nodes)
    adj = np.zeros((n, n))
    for st, ed, w in di_graph.edges(data='weight', default=1):
        adj[st, ed] = w
    return adj


def transform_adj_to_DiGraph(adj):
    n = adj.shape[0]
    di_graph = nx.DiGraph()
    di_graph.add_nodes_from(range(n))
    for i in range(n):
        for j in range(n):
            if(i != j):
                if(adj[i, j] > 0):
                    di_graph.add_edge(i, j, weight=adj[i, j])
    return di_graph


def get_lcc(di_graph):
    di_graph = max(nx.weakly_connected_component_subgraphs(di_graph), key=len)
    tdl_nodes = list(di_graph.nodes())
    nodeListMap = dict(zip(tdl_nodes, range(len(tdl_nodes))))
    nx.relabel_nodes(di_graph, nodeListMap, copy=False)
    return di_graph, nodeListMap


def print_graph_stats(G):
    print('# of nodes: %d, # of edges: %d' % (len(G.nodes),
                                              len(G.edges)))


def sample_graph(di_graph, n_sampled_nodes=None):
    node_num = len(di_graph.nodes)
    if n_sampled_nodes and node_num > n_sampled_nodes:
        node_l = np.random.choice(node_num, n_sampled_nodes, replace=False)
        node_l_inv = {v: k for k, v in enumerate(node_l)}
        sampled_graph = nx.DiGraph()
        sampled_graph.add_nodes_from(range(n_sampled_nodes))
        for st, ed, w in di_graph.edges(data='weight', default=1):
            try:
                v_i = node_l_inv[st]
                v_j = node_l_inv[ed]
                sampled_graph.add_edge(v_i, v_j, weight=w)
            except:
                continue
        return sampled_graph, node_l
    else:
        return di_graph, np.arange(len(di_graph.nodes))


def randwalk_DiGraph_to_adj(di_graph, node_frac=0.1,
                            n_walks_per_node=5, len_rw=2):
    t0 = time.time()
    n = len(di_graph.nodes)
    adj = np.zeros((n, n))
    rw_node_num = int(node_frac * n)
    rw_node_list = np.random.choice(n, size=[rw_node_num],
                                    replace=False, p=None)
    for node in rw_node_list:
        for walk in range(n_walks_per_node):
            cur_node = node
            for step in range(len_rw):
                cur_neighbors = di_graph.neighbors(cur_node)
                try:
                    neighbor_node = np.random.choice(cur_neighbors)
                except:
                    continue
                try:
                    adj[cur_node, neighbor_node] = di_graph.get_edge_data(
                        cur_node,
                        neighbor_node
                    )['weight']
                    adj[neighbor_node, cur_node] = di_graph.get_edge_data(
                        cur_node,
                        neighbor_node
                    )['weight']
                except KeyError:
                    adj[cur_node, neighbor_node] = 1
                    adj[neighbor_node, cur_node] = 1
                cur_node = neighbor_node
    print('Time taken for random walk  on {0} nodes = {1}'.format(n, time.time() - t0))
    return adj


def addChaos(di_graphs, k):
    anomaly_time_steps = sorted(random.sample(range(len(di_graphs)), k))
    for t in anomaly_time_steps:
        n = len(di_graphs[t].nodes)
        e = len(di_graphs[t].edges)
        di_graphs[t] = nx.fast_gnp_random_graph(n, e / float(n * (n - 1)),
                                                seed=None, directed=False)
        di_graphs[t] = di_graphs[t].to_directed()
    return di_graphs, anomaly_time_steps


def addNodeAnomalies(di_graphs, p, k):
    anomaly_time_steps = sorted(random.sample(range(len(di_graphs)), k))
    for t in anomaly_time_steps:
        n_nodes = len(di_graphs[t].nodes)
        anomalous_nodes_idx = np.random.choice([0, 1],
                                               size=(n_nodes, 1),
                                               p=(1 - p, p))
        node_list = np.array(list(di_graphs[t].nodes()))
        node_list = node_list.reshape((n_nodes, 1))
        anomalous_nodes = np.multiply(anomalous_nodes_idx, node_list)
        anomalous_nodes = anomalous_nodes[anomalous_nodes > 0]
        # pdb.set_trace()
        di_graphs[t].add_edges_from(
            itertools.product(list(anomalous_nodes), range(n_nodes))
        )
        di_graphs[t].add_edges_from(
            itertools.product(range(n_nodes), list(anomalous_nodes))
        )
        print('Nodes: %d, Edges: %d' % (len(di_graphs[t].nodes),
                                        len(di_graphs[t].edges)))
    return anomaly_time_steps


def saveGraphToEdgeListTxt(graph, file_name):
    with open(file_name, 'w') as f:
        f.write('%d\n' % len(graph.nodes))
        f.write('%d\n' % len(graph.edges))
        for i, j, w in graph.edges(data='weight', default=1):
            f.write('%d %d %f\n' % (i, j, w))


def saveGraphToEdgeListTxtn2v(graph, file_name):
    with open(file_name, 'w') as f:
        for i, j, w in graph.edges(data='weight', default=1):
            f.write('%d %d %f\n' % (i, j, w))


def loadGraphFromEdgeListTxt(file_name, directed=True):
    with open(file_name, 'r') as f:
        # n_nodes = f.readline()
        # f.readline() # Discard the number of edges
        if directed:
            G = nx.DiGraph()
        else:
            G = nx.Graph()
        for line in f:
            edge = line.strip().split()
            if len(edge) == 3:
                w = float(edge[2])
            else:
                w = 1.0
            G.add_edge(int(edge[0]), int(edge[1]), weight=w)
    return G


def loadEmbedding(file_name):
    with open(file_name, 'r') as f:
        n, d = f.readline().strip().split()
        X = np.zeros((int(n), int(d)))
        for line in f:
            emb = line.strip().split()
            emb_fl = [float(emb_i) for emb_i in emb[1:]]
            X[int(emb[0]), :] = emb_fl
    return X


def loadSBMGraph(file_prefix):
    graph_file = file_prefix + '_graph.gpickle'
    G = nx.read_gpickle(graph_file)
    node_file = file_prefix + '_node.pkl'
    with open(node_file, 'rb') as fp:
        node_community = pickle.load(fp)
    return (G, node_community)


def loadRealGraphSeries(file_prefix, startId, endId):
    graphs = []
    for file_id in range(startId, endId + 1):
        graph_file = file_prefix + str(file_id) + '_graph.gpickle'
        graphs.append(nx.read_gpickle(graph_file))
    return graphs


def saveRealGraphSeries(G, file_prefix='graphs/day_'):
    for idx in range(len(G)):
        f_name = file_prefix + str(idx) + "_graph.gpickle"
        # cPickle.dump(G[idx], open(f_name, 'wb'))
        nx.write_gpickle(G[idx], f_name)


def loadDynamicSBmGraph(file_perfix, length):
    graph_files = ['%s_%d_graph.gpickle' % (file_perfix, i) for i in range(length)]
    info_files = ['%s_%d_node.pkl' % (file_perfix, i) for i in range(length)]

    graphs = [nx.read_gpickle(graph_file) for graph_file in graph_files]

    nodes_comunities = []
    perturbations = []
    for info_file in info_files:
        with open(info_file, 'rb') as fp:
            node_infos = pickle.load(fp)
            nodes_comunities.append(node_infos['community'])
            perturbations.append(node_infos['perturbation'])

    return zip(graphs, nodes_comunities, perturbations)


def saveDynamicSBmGraph(file_perfix, dynamic_graphs):
    length = len(dynamic_graphs)
    graph_files = ['%s_%d_graph.gpickle' % (file_perfix, i) for i in range(length)]
    info_files = ['%s_%d_node.pkl' % (file_perfix, i) for i in range(length)]

    for i in range(length):
        # save graph
        nx.write_gpickle(dynamic_graphs[i][0], graph_files[i])
        # save additional node info
        with open(info_files[i], 'wb') as fp:
            node_infos = {}
            node_infos['community'] = dynamic_graphs[i][1]
            node_infos['perturbation'] = dynamic_graphs[i][2]
            pickle.dump(node_infos, fp)



from abc import ABCMeta, ABC, abstractmethod
import numpy as np


class StaticGraphEmbedding(ABC):
    __metaclass__ = ABCMeta

    def __init__(self, *args, **kwargs):
        """Initialize the Embedding class
        """
        self._method_name = None
        self._d = None
        self._X = None
        self.hyper_params.update(kwargs)
        for key in self.hyper_params.keys():
            self.__setattr__('_%s' % key, self.hyper_params[key])
        for dictionary in args:
            for key in dictionary:
                self.__setattr__('_%s' % key, dictionary[key])

    def get_method_name(self):
        """ Returns the name for the embedding method

        Return:
            The name of embedding
        """
        return self._method_name

    def get_method_summary(self):
        """ Returns the summary for the embedding include method name and paramater setting

        Return:
            A summary string of the method
        """

        return '%s_%d' % (self._method_name, self._d)

    def get_embedding(self):
        """ Returns the learnt embedding

        Return:
            A numpy array of size #nodes * d
        """
        if self._X is None:
            raise ValueError("Embedding not learned yet")
        return self._X

    def get_reconstructed_adj(self, X=None, node_l=None):
        """Compute the adjacency matrix from the learned embedding

        Returns:
            A numpy array of size #nodes * #nodes containing the reconstructed adjacency matrix.
        """
        if X is not None:
            node_num = X.shape[0]
            self._X = X
        else:
            node_num = self._node_num
        adj_mtx_r = np.zeros((node_num, node_num))
        for v_i in range(node_num):
            for v_j in range(node_num):
                if v_i == v_j:
                    continue
                adj_mtx_r[v_i, v_j] = self.get_edge_weight(v_i, v_j)
        return adj_mtx_r

    @abstractmethod
    def learn_embedding(self, graph):
        """Learning the graph embedding from the adjcency matrix.

        Args:
            graph: the graph to embed in networkx DiGraph format
        """

    @abstractmethod
    def get_edge_weight(self, i, j):
        """Compute the weight for edge between node i and node j

        Args:
            i, j: two node id in the graph for embedding
        Returns:
            A single number represent the weight of edge between node i and node j

        """


"""
Run the graph embedding methods on Karate graph and evaluate them on
graph reconstruction and visualization. Please copy the
gem/data/karate.edgelist to the working directory
"""
import os
import unittest

import numpy as np

from gem.utils import graph_util

from gem.embedding.gf import GraphFactorization
from gem.embedding.hope import HOPE
from gem.embedding.lap import LaplacianEigenmaps
from gem.embedding.lle import LocallyLinearEmbedding
from gem.embedding.node2vec import node2vec
from gem.embedding.sdne import SDNE

from tests.fit_model import fit_model


class KarateTest(unittest.TestCase):

    def setUp(self) -> None:
        # File that contains the edges. Format: source target
        # Optionally, you can add weights as third column: source target weight
        self.source_dir = os.path.dirname(os.path.abspath(__file__))
        edge_f = os.path.join(self.source_dir, 'data/karate.edgelist')
        # Specify whether the edges are directed
        isDirected = True

        # Load graph
        G = graph_util.loadGraphFromEdgeListTxt(edge_f, directed=isDirected)
        self.G = G.to_directed()

    def test_GraphFactorization(self):
        model = GraphFactorization(d=2, max_iter=50000, eta=1 * 10**-4, regu=1.0, data_set='karate')
        target = np.loadtxt(os.path.join(self.source_dir, 'karate_res/GraphFactorization.txt'))
        self.internal_model_test(model, target, mae_close=True)

    def test_HOPE(self):
        model = HOPE(d=4, beta=0.01)
        target = np.loadtxt(os.path.join(self.source_dir, 'karate_res/HOPE.txt'))
        self.internal_model_test(model, target)

    def test_LaplacianEigenmaps(self):
        model = LaplacianEigenmaps(d=2)
        target = np.loadtxt(os.path.join(self.source_dir, 'karate_res/LaplacianEigenmaps.txt'))
        self.internal_model_test(model, target)

    def test_LocallyLinearEmbedding(self):
        model = LocallyLinearEmbedding(d=2)
        target = np.loadtxt(os.path.join(self.source_dir, 'karate_res/LocallyLinearEmbedding.txt'))
        self.internal_model_test(model, target, mae_close=True)

    def test_node2vec(self):
        model = node2vec(d=2, max_iter=1, walk_len=80, num_walks=10, con_size=10, ret_p=1, inout_p=1)
        target = np.loadtxt(os.path.join(self.source_dir, 'karate_res/node2vec.txt'))
        self.internal_model_test(model, target, mae_close=True)

    def test_SDNE(self):
        model = SDNE(d=2, beta=5, alpha=1e-5, nu1=1e-6, nu2=1e-6, K=3, n_units=[50, 15,], rho=0.3, n_iter=50,
                     xeta=0.01, n_batch=100, modelfile=['enc_model.json', 'dec_model.json'],
                     weightfile=['enc_weights.hdf5', 'dec_weights.hdf5'])
        target = np.loadtxt(os.path.join(self.source_dir, 'karate_res/SDNE.txt'))
        self.internal_model_test(model, target, mae_close=True)

    def internal_model_test(self, model, target, verbose: bool = False, mae_close: bool = False):
        MAP, prec_curv, err, err_baseline = fit_model(self.G, model)
        # ---------------------------------------------------------------------------------
        if verbose:
            print(("\tMAP: {} \t preccision curve: {}\n\n\n\n" + '-' * 100).format(MAP, prec_curv[:5]))
        # ---------------------------------------------------------------------------------
        if not mae_close:
            self.assertTrue(np.allclose(model.get_embedding(), target))
        else:
            self.assertTrue(abs(np.mean(target-model.get_embedding())) < .3)
